import ssl
from concurrent.futures import ThreadPoolExecutor
from urllib.parse import urlparse

from tornado.tcpclient import TCPClient
from tornado.iostream import StreamClosedError
from tornado import netutil

from distributed.comm.core import Connector
from distributed.comm.registry import Backend, backends
from distributed.comm.tcp import (
    TLS,
    MAX_BUFFER_SIZE,
    get_stream_address,
    convert_stream_closed_error,
)
from distributed.utils import ensure_ip, get_ip


def parse_gateway_address(address):
    if not address.startswith("gateway://"):
        address = "gateway://" + address
    parsed = urlparse(address)
    if not parsed.path:
        raise ValueError("Gateway address %r missing path component" % address)
    path = parsed.path.strip("/")
    return parsed.hostname, parsed.port, path


class GatewayConnector(Connector):
    _executor = ThreadPoolExecutor(2)
    _resolver = netutil.ExecutorResolver(close_executor=False, executor=_executor)
    client = TCPClient(resolver=_resolver)

    async def connect(self, address, deserialize=True, **connection_args):
        ip, port, sni = parse_gateway_address(address)
        ctx = connection_args.get("ssl_context")
        if not isinstance(ctx, ssl.SSLContext):
            raise TypeError(
                "Gateway expects a `ssl_context` argument of type "
                "ssl.SSLContext, instead got %s" % ctx
            )

        try:
            plain_stream = await self.client.connect(
                ip, port, max_buffer_size=MAX_BUFFER_SIZE
            )
            stream = await plain_stream.start_tls(
                False, ssl_options=ctx, server_hostname=sni
            )
            if stream.closed() and stream.error:
                raise StreamClosedError(stream.error)

        except StreamClosedError as e:
            # The socket connect() call failed
            convert_stream_closed_error(self, e)

        local_address = "tls://" + get_stream_address(stream)
        peer_address = "gateway://" + address
        return TLS(stream, local_address, peer_address, deserialize)


class GatewayBackend(Backend):
    # I/O
    def get_connector(self):
        return GatewayConnector()

    def get_listener(self, *args, **kwargs):
        raise NotImplementedError("Listening on a gateway connection")

    # Address handling
    def get_address_host(self, loc):
        return parse_gateway_address(loc)[0]

    def get_address_host_port(self, loc):
        return parse_gateway_address(loc)[:2]

    def resolve_address(self, loc):
        host, port, path = parse_gateway_address(loc)
        host = ensure_ip(host)
        return "%s:%d/%s" % (host, port, path)

    def get_local_address_for(self, loc):
        host, port, path = parse_gateway_address(loc)
        host = ensure_ip(host)
        host = get_ip(host)
        return "%s:%d/%s" % (host, port, path)


backends["gateway"] = GatewayBackend()



from .client import (
    Gateway,
    GatewayCluster,
    GatewayClusterError,
    GatewayServerError,
    GatewayWarning,
)
from .auth import KerberosAuth, BasicAuth
from .options import Options

# Load configuration
from . import config

del config

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



import weakref
from collections import OrderedDict
from collections.abc import MutableMapping, Sequence
from keyword import iskeyword


__all__ = ("Options",)


class Options(MutableMapping):
    """Options to submit to dask-gateway.

    A mutable-mapping that describes all cluster options available for user's
    to set. Options can be modified programatically, or via a widget when using
    a web interface (e.g. Jupyter Notebooks) with ``ipywidgets`` installed.

    Examples
    --------
    Options objects are normally created via the ``Gateway.cluster_options``
    method:

    >>> options = gateway.cluster_options()  # doctest: +SKIP

    Available options can then be accessed or set via attribute or key:

    >>> options.worker_cores  # doctest: +SKIP
    1
    >>> options["worker_cores"]  # doctest: +SKIP
    1
    >>> options.worker_cores = 2  # doctest: +SKIP
    >>> options.worker_cores  # doctest: +SKIP
    2

    Accessing invalid options error appropriately:

    >>> options.not_a_valid_option = 'myvalue'  # doctest: +SKIP
    Traceback (most recent call last):
        ...
    AttributeError: No option 'not_a_valid_option' available

    Errors are also raised if values being set are invalid (e.g. wrong type,
    out-of-bounds, etc...).
    """

    def __init__(self, *fields):
        object.__setattr__(self, "_fields", OrderedDict((f.field, f) for f in fields))

    @classmethod
    def _from_spec(cls, spec):
        return cls(*(Field._from_spec(s) for s in spec))

    def _widget(self):
        if not hasattr(self, "_cached_widget"):
            try:
                import ipywidgets

                children = [ipywidgets.HTML("<h2>Cluster Options</h2>")]
                children.extend([f.widget() for f in self._fields.values()])
                column = ipywidgets.Box(
                    children=children,
                    layout=ipywidgets.Layout(
                        display="flex", flex_flow="column", align_items="stretch"
                    ),
                )
                widget = ipywidgets.Box(children=[column])
            except ImportError:
                widget = None
            object.__setattr__(self, "_cached_widget", widget)
        return self._cached_widget

    def _ipython_display_(self, **kwargs):
        widget = self._widget()
        if widget is not None:
            return widget._ipython_display_(**kwargs)
        from IPython.lib.pretty import pprint

        pprint(self)

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text("Options<...>")
        else:
            with p.group(8, "Options<", ">"):
                for idx, field in enumerate(self._fields.values()):
                    if idx:
                        p.text(",")
                        p.breakable()
                    p.text("%s=" % field.field)
                    p.pretty(field.value)

    def _get(self, key, exc_cls):
        try:
            return self._fields[key].get()
        except KeyError:
            raise exc_cls("No option %r available" % key) from None

    def _set(self, key, value, exc_cls):
        try:
            self._fields[key].set(value)
        except KeyError:
            raise exc_cls("No option %r available" % key) from None

    def __getattr__(self, key):
        return self._get(key, AttributeError)

    def __setattr__(self, key, value):
        return self._set(key, value, AttributeError)

    def __getitem__(self, key):
        return self._get(key, KeyError)

    def __setitem__(self, key, value):
        return self._set(key, value, KeyError)

    def __delitem__(self, key):
        raise TypeError("Cannot delete fields")

    __delattr__ = __delitem__

    def __iter__(self):
        return iter(self._fields)

    def __len__(self):
        return len(self._fields)

    def __dir__(self):
        o = set(dir(type(self)))
        o.update(self.__dict__)
        o.update(f for f in self._fields if f.isidentifier() and not iskeyword(f))
        return list(o)


def register_field_type(name):
    def inner(cls):
        Field._field_types[name] = cls
        return cls

    return inner


class Field(object):
    """A single option field"""

    _field_types = {}

    def __init__(self, field, default, label=None):
        self.field = field
        self.value = self.validate(default)
        self.label = label or field

        self._widgets = weakref.WeakSet()

    @classmethod
    def _from_spec(cls, spec):
        field = spec["field"]
        default = spec["default"]
        label = spec["label"]
        type_spec = dict(spec["spec"])
        typ = type_spec.pop("type")
        return cls._field_types[typ](field, default, label=label, **type_spec)

    def validate(self, x):
        raise NotImplementedError

    def set(self, value):
        self.value = self.validate(value)
        # Update all linked widgets
        for w in self._widgets:
            w.value = self.value

    def get(self):
        return self.value

    def widget(self):
        import ipywidgets

        def handler(change):
            self.set(change.new)

        input = self._widget()
        input.observe(handler, "value")
        self._widgets.add(input)

        label = ipywidgets.HTML(
            "<p style='font-weight: bold; margin-right: 8px'>%s:</p>" % self.label
        )

        row = ipywidgets.Box(
            children=[label, input],
            layout=ipywidgets.Layout(
                display="flex", flex_flow="row wrap", justify_content="space-between"
            ),
        )

        return row


@register_field_type("string")
class String(Field):
    """A string option field"""

    def validate(self, x):
        if not isinstance(x, str):
            raise TypeError("%s must be a string, got %r" % (self.field, x))
        return x

    def _widget(self):
        import ipywidgets

        return ipywidgets.Text(value=self.value, continuous_update=False)


@register_field_type("bool")
class Bool(Field):
    """A boolean option field"""

    def validate(self, x):
        if not isinstance(x, bool):
            raise TypeError("%s must be a bool, got %r" % (self.field, x))
        return x

    def _widget(self):
        import ipywidgets

        return ipywidgets.Checkbox(value=self.value, indent=False)


class Number(Field):
    def __init__(self, field, default, label=None, min=None, max=None):
        # Temporarily set to allow `validate` to work
        self.min = self.max = None
        if min is not None:
            self.min = self.validate(min)
        if max is not None:
            self.max = self.validate(max)
        super().__init__(field, default, label=label)

    def validate(self, x):
        if self.min is not None and x < self.min:
            raise ValueError("%s must be >= %f, got %s" % (self.field, self.min, x))
        if self.max is not None and x > self.max:
            raise ValueError("%s must be <= %f, got %s" % (self.field, self.max, x))
        return x


@register_field_type("int")
class Integer(Number):
    """An integer option field"""

    def validate(self, x):
        if not isinstance(x, int):
            raise TypeError("%s must be an integer, got %r" % (self.field, x))
        return super().validate(x)

    def _widget(self):
        import ipywidgets

        if self.min is None and self.max is None:
            return ipywidgets.IntText(value=self.value)
        else:
            return ipywidgets.BoundedIntText(
                value=self.value,
                min=self.min or -(2 ** 63),
                max=self.max or (2 ** 63 - 1),
            )


@register_field_type("float")
class Float(Number):
    """A float option field"""

    def validate(self, x):
        if isinstance(x, int):
            x = float(x)
        if not isinstance(x, float):
            raise TypeError("%s must be a float, got %r" % (self.field, x))
        return super().validate(x)

    def _widget(self):
        import ipywidgets

        if self.min is None and self.max is None:
            return ipywidgets.FloatText(value=self.value)
        else:
            return ipywidgets.BoundedFloatText(
                value=self.value,
                min=self.min or float("-inf"),
                max=self.max or float("inf"),
            )


@register_field_type("select")
class Select(Field):
    """A select option field"""

    def __init__(self, field, default, label=None, options=None):
        if not isinstance(options, Sequence):
            raise TypeError("options must be a sequence")
        elif not len(options):
            raise ValueError("There must be at least one option")
        for o in options:
            if not isinstance(o, str):
                raise TypeError("Select options must be strings, got %r" % o)
        self.options = tuple(options)
        self._options_set = set(options)
        super().__init__(field, default, label=label)

    def validate(self, x):
        if not isinstance(x, str):
            raise TypeError("%s must be a string, got %r" % (self.field, x))
        if x not in self._options_set:
            raise ValueError("%r is not a valid option for %s" % (x, self.field))
        return x

    def _widget(self):
        import ipywidgets

        return ipywidgets.Dropdown(value=self.value, options=self.options)



import enum
import json
import os
import ssl
import tempfile
import warnings
import weakref
from datetime import timedelta, datetime
from threading import get_ident
from urllib.parse import urlparse

import dask
from distributed import Client
from distributed.security import Security
from distributed.utils import LoopRunner, sync, thread_state, format_bytes, log_errors
from tornado import gen
from tornado.httpclient import AsyncHTTPClient, HTTPRequest, HTTPError
from tornado.httputil import HTTPHeaders

from .cookiejar import CookieJar

# Register gateway protocol
from . import comm
from .auth import get_auth
from .options import Options

del comm

__all__ = ("Gateway", "GatewayCluster", "GatewayClusterError", "GatewayServerError")


class GatewayClusterError(Exception):
    """Exception related to starting/stopping/scaling of a gateway cluster"""


class GatewayServerError(Exception):
    """Exception related to the operation of the gateway server.

    Indicates an internal error in the gateway server.
    """


class GatewayWarning(UserWarning):
    """Warnings raised by the Gateway client"""


class GatewaySecurity(Security):
    """A security implementation that temporarily stores credentials on disk.

    The normal ``Security`` class assumes credentials already exist on disk,
    but our credentials exist only in memory. Since Python's SSLContext doesn't
    support directly loading credentials from memory, we write them temporarily
    to disk when creating the context, then delete them immediately."""

    def __init__(self, tls_key, tls_cert):
        self.tls_key = tls_key
        self.tls_cert = tls_cert

    def __repr__(self):
        return "GatewaySecurity<...>"

    def get_connection_args(self, role):
        with tempfile.TemporaryDirectory() as tempdir:
            key_path = os.path.join(tempdir, "dask.pem")
            cert_path = os.path.join(tempdir, "dask.crt")
            with open(key_path, "w") as f:
                f.write(self.tls_key)
            with open(cert_path, "w") as f:
                f.write(self.tls_cert)
            ctx = ssl.create_default_context(
                purpose=ssl.Purpose.SERVER_AUTH, cafile=cert_path
            )
            ctx.verify_mode = ssl.CERT_REQUIRED
            ctx.check_hostname = False
            ctx.load_cert_chain(cert_path, key_path)
        return {"ssl_context": ctx, "require_encryption": True}


class _IntEnum(enum.IntEnum):
    @classmethod
    def _create(cls, x):
        return x if isinstance(x, cls) else cls.from_name(x)

    @classmethod
    def from_name(cls, name):
        """Create an enum value from a name"""
        try:
            return cls[name.upper()]
        except KeyError:
            pass
        raise ValueError("%r is not a valid %s" % (name, cls.__name__))


class ClusterStatus(_IntEnum):
    """The cluster's status.

    Attributes
    ----------
    STARTING : ClusterStatus
        The cluster is starting.
    STARTED : ClusterStatus
        The cluster has started, but hasn't connected to the gateway.
    RUNNING : ClusterStatus
        The cluster is running.
    STOPPING : ClusterStatus
        The cluster is stopping, but not fully stopped.
    STOPPED : ClusterStatus
        The cluster was shutdown by user or admin request.
    FAILED : ClusterStatus
        A failure occurred during any of the above states, see the logs for
        more information.
    """

    STARTING = 1
    STARTED = 2
    RUNNING = 3
    STOPPING = 4
    STOPPED = 5
    FAILED = 6


class ClusterReport(object):
    """A report on a cluster's state.

    Attributes
    ----------
    name : str
        The cluster's name, a unique id.
    options : dict
        User provided configuration for this cluster.
    status : ClusterStatus
        The cluster's status.
    scheduler_address : str or None
        The address of the scheduler, or None if not currently running.
    dashboard_link : str or None
        A link to the dashboard, or None if not currently running.
    start_time : datetime.datetime
        The time the cluster was started.
    stop_time : datetime.datetime or None
        The time the cluster was stopped, or None if currently running.
    tls_cert : str or None
        The tls certificate. None if not currently running or a full report
        wasn't requested.
    tls_key : str or None
        The tls key. None if not currently running or a full report wasn't
        requested.
    """

    __slots__ = (
        "name",
        "options",
        "status",
        "scheduler_address",
        "dashboard_link",
        "start_time",
        "stop_time",
        "tls_cert",
        "tls_key",
    )

    def __init__(
        self,
        name,
        options,
        status,
        scheduler_address,
        dashboard_link,
        start_time,
        stop_time,
        tls_cert=None,
        tls_key=None,
    ):
        self.name = name
        self.options = options
        self.status = status
        self.scheduler_address = scheduler_address
        self.dashboard_link = dashboard_link
        self.start_time = start_time
        self.stop_time = stop_time
        self.tls_cert = tls_cert
        self.tls_key = tls_key

    def __repr__(self):
        return "ClusterReport<name=%s, status=%s>" % (self.name, self.status.name)

    @property
    def security(self):
        """A security object for this cluster, if present, else None"""
        return (
            None
            if self.tls_key is None
            else GatewaySecurity(tls_key=self.tls_key, tls_cert=self.tls_cert)
        )

    @classmethod
    def _from_json(cls, gateway_address, proxy_address, msg):
        start_time = datetime.fromtimestamp(msg.pop("start_time") / 1000)
        stop_time = msg.pop("stop_time")
        if stop_time is not None:
            stop_time = datetime.fromtimestamp(stop_time / 1000)

        name = msg.pop("name")
        status = ClusterStatus._create(msg.pop("status"))
        scheduler_address = (
            "%s/%s" % (proxy_address, name) if status == ClusterStatus.RUNNING else None
        )

        dashboard_route = msg.pop("dashboard_route")
        dashboard_link = (
            "%s%s" % (gateway_address, dashboard_route) if dashboard_route else None
        )

        return cls(
            start_time=start_time,
            stop_time=stop_time,
            name=name,
            status=status,
            scheduler_address=scheduler_address,
            dashboard_link=dashboard_link,
            **msg,
        )


class Gateway(object):
    """A client for a Dask Gateway Server.

    Parameters
    ----------
    address : str, optional
        The address to the gateway server.
    proxy_address : str, int, optional
        The address of the scheduler proxy server. If an int, it's used as the
        port, with the host/ip taken from ``address``. Provide a full address
        if a different host/ip should be used.
    auth : GatewayAuth, optional
        The authentication method to use.
    asynchronous : bool, optional
        If true, starts the client in asynchronous mode, where it can be used
        in other async code.
    loop : IOLoop, optional
        The IOLoop instance to use. Defaults to the current loop in
        asynchronous mode, otherwise a background loop is started.
    """

    def __init__(
        self, address=None, proxy_address=None, auth=None, asynchronous=False, loop=None
    ):
        if address is None:
            address = dask.config.get("gateway.address")
        if address is None:
            raise ValueError(
                "No dask-gateway address provided or found in configuration"
            )
        address = address.rstrip("/")

        if proxy_address is None:
            proxy_address = dask.config.get("gateway.proxy-address")
        if proxy_address is None:
            raise ValueError(
                "No dask-gateway proxy address provided or found in configuration"
            )
        if isinstance(proxy_address, int):
            parsed = urlparse(address)
            proxy_netloc = "%s:%d" % (parsed.hostname, proxy_address)
        elif isinstance(proxy_address, str):
            parsed = urlparse(proxy_address)
            proxy_netloc = parsed.netloc if parsed.netloc else proxy_address
        proxy_address = "gateway://%s" % proxy_netloc

        self.address = address
        self.proxy_address = proxy_address

        self._auth = get_auth(auth)
        self._cookie_jar = CookieJar()

        self._asynchronous = asynchronous
        self._loop_runner = LoopRunner(loop=loop, asynchronous=asynchronous)
        self.loop = self._loop_runner.loop

        self._loop_runner.start()
        if self._asynchronous:
            self._started = self._start()
        else:
            self.sync(self._start)

    async def _start(self):
        self._http_client = AsyncHTTPClient()

    def close(self):
        """Close this gateway client"""
        if not self.asynchronous:
            self._loop_runner.stop()

    def __del__(self):
        # __del__ is still called, even if __init__ failed. Only close if init
        # actually succeeded.
        if hasattr(self, "_started"):
            self.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def __await__(self):
        return self._started.__await__()

    async def __aenter__(self):
        await self._started
        return self

    async def __aexit__(self, typ, value, traceback):
        pass

    def __repr__(self):
        return "Gateway<%s>" % self.address

    @property
    def asynchronous(self):
        return (
            self._asynchronous
            or getattr(thread_state, "asynchronous", False)
            or (
                hasattr(self.loop, "_thread_identity")
                and self.loop._thread_identity == get_ident()
            )
        )

    def sync(self, func, *args, **kwargs):
        if kwargs.pop("asynchronous", None) or self.asynchronous:
            callback_timeout = kwargs.pop("callback_timeout", None)
            future = func(*args, **kwargs)
            if callback_timeout is not None:
                future = gen.with_timeout(timedelta(seconds=callback_timeout), future)
            return future
        else:
            return sync(self.loop, func, *args, **kwargs)

    async def _fetch(self, req):
        try:
            self._cookie_jar.pre_request(req)
            resp = await self._http_client.fetch(req, raise_error=False)
            if resp.code == 401:
                context = self._auth.pre_request(req, resp)
                resp = await self._http_client.fetch(req, raise_error=False)
                self._auth.post_response(req, resp, context)
            self._cookie_jar.post_response(resp)
            if resp.error:
                if resp.code == 599:
                    raise TimeoutError("Request timed out")
                else:
                    try:
                        msg = json.loads(resp.body)["error"]
                    except Exception:
                        msg = resp.body.decode()

                    if resp.code in {404, 422}:
                        raise ValueError(msg)
                    elif resp.code == 409:
                        raise GatewayClusterError(msg)
                    elif resp.code == 500:
                        raise GatewayServerError(msg)
                    else:
                        resp.rethrow()
        except HTTPError as exc:
            # Tornado 6 still raises these above with raise_error=False
            if exc.code == 599:
                raise TimeoutError("Request timed out")
            # Should never get here!
            raise
        return resp

    async def _clusters(self, status=None):
        if status is not None:
            if isinstance(status, (str, ClusterStatus)):
                status = [ClusterStatus._create(status)]
            else:
                status = [ClusterStatus._create(s) for s in status]
            query = "?status=" + ",".join(s.name for s in status)
        else:
            query = ""

        url = "%s/gateway/api/clusters/%s" % (self.address, query)
        req = HTTPRequest(url=url)
        resp = await self._fetch(req)
        return [
            ClusterReport._from_json(self.address, self.proxy_address, r)
            for r in json.loads(resp.body).values()
        ]

    def list_clusters(self, status=None, **kwargs):
        """List clusters for this user.

        Parameters
        ----------
        status : ClusterStatus, str, or list, optional
            The cluster status (or statuses) to select. Valid options are
            'starting', 'started', 'running', 'stopping', 'stopped', 'failed'.
            By default selects active clusters ('starting', 'started',
            'running').

        Returns
        -------
        clusters : list of ClusterReport
        """
        return self.sync(self._clusters, status=status, **kwargs)

    async def _cluster_options(self):
        url = "%s/gateway/api/clusters/options" % self.address
        req = HTTPRequest(url=url, method="GET")
        resp = await self._fetch(req)
        data = json.loads(resp.body)
        return Options._from_spec(data["cluster_options"])

    def cluster_options(self, **kwargs):
        """Get the available cluster configuration options.

        Returns
        -------
        cluster_options : Options
            A dict of cluster options.
        """
        return self.sync(self._cluster_options, **kwargs)

    async def _submit(self, cluster_options=None, **kwargs):
        url = "%s/gateway/api/clusters/" % self.address
        if cluster_options is not None:
            if not isinstance(cluster_options, Options):
                raise TypeError(
                    "cluster_options must be an `Options`, got %r"
                    % type(cluster_options).__name__
                )
            options = dict(cluster_options)
            options.update(kwargs)
        else:
            options = kwargs
        req = HTTPRequest(
            url=url,
            method="POST",
            body=json.dumps({"cluster_options": options}),
            headers=HTTPHeaders({"Content-type": "application/json"}),
        )
        resp = await self._fetch(req)
        data = json.loads(resp.body)
        return data["name"]

    def submit(self, cluster_options=None, **kwargs):
        """Submit a new cluster to be started.

        This returns quickly with a ``cluster_name``, which can later be used
        to connect to the cluster.

        Parameters
        ----------
        cluster_options : Options, optional
            An ``Options`` object describing the desired cluster configuration.
        **kwargs :
            Additional cluster configuration options. If ``cluster_options`` is
            provided, these are applied afterwards as overrides. Available
            options are specific to each deployment of dask-gateway, see
            ``cluster_options`` for more information.

        Returns
        -------
        cluster_name : str
            The cluster name.
        """
        return self.sync(self._submit, **kwargs)

    async def _cluster_report(self, cluster_name, wait=False):
        params = "?wait" if wait else ""
        url = "%s/gateway/api/clusters/%s%s" % (self.address, cluster_name, params)
        req = HTTPRequest(url=url)
        resp = await self._fetch(req)
        return ClusterReport._from_json(
            self.address, self.proxy_address, json.loads(resp.body)
        )

    async def _connect(self, cluster_name):
        while True:
            try:
                report = await self._cluster_report(cluster_name, wait=True)
            except TimeoutError:
                # Timeout, ignore
                pass
            else:
                if report.status is ClusterStatus.RUNNING:
                    return GatewayCluster(
                        gateway=self,
                        name=report.name,
                        scheduler_address=report.scheduler_address,
                        dashboard_link=report.dashboard_link,
                        security=report.security,
                    )
                elif report.status is ClusterStatus.FAILED:
                    raise GatewayClusterError(
                        "Cluster %r failed to start, see logs for "
                        "more information" % cluster_name
                    )
                elif report.status is ClusterStatus.STOPPED:
                    raise GatewayClusterError(
                        "Cluster %r is already stopped" % cluster_name
                    )
            # Not started yet, try again later
            await gen.sleep(0.5)

    def connect(self, cluster_name, **kwargs):
        """Connect to a submitted cluster.

        Returns
        -------
        cluster : GatewayCluster
        """
        return self.sync(self._connect, cluster_name, **kwargs)

    async def _new_cluster(self, **kwargs):
        cluster_name = await self._submit(**kwargs)
        try:
            return await self._connect(cluster_name)
        except GatewayClusterError:
            raise
        except BaseException:
            # Ensure cluster is stopped on error
            await self._stop_cluster(cluster_name)
            raise

    def new_cluster(self, cluster_options=None, **kwargs):
        """Submit a new cluster to the gateway, and wait for it to be started.

        Same as calling ``submit`` and ``connect`` in one go.

        Parameters
        ----------
        cluster_options : Options, optional
            An ``Options`` object describing the desired cluster configuration.
        **kwargs :
            Additional cluster configuration options. If ``cluster_options`` is
            provided, these are applied afterwards as overrides. Available
            options are specific to each deployment of dask-gateway, see
            ``cluster_options`` for more information.

        Returns
        -------
        cluster : GatewayCluster
        """
        return self.sync(self._new_cluster, cluster_options=cluster_options, **kwargs)

    async def _stop_cluster(self, cluster_name):
        url = "%s/gateway/api/clusters/%s" % (self.address, cluster_name)
        req = HTTPRequest(url=url, method="DELETE")
        await self._fetch(req)

    def stop_cluster(self, cluster_name, **kwargs):
        """Stop a cluster.

        Parameters
        ----------
        cluster_name : str
            The cluster name.
        """
        return self.sync(self._stop_cluster, cluster_name, **kwargs)

    async def _scale_cluster(self, cluster_name, n):
        url = "%s/gateway/api/clusters/%s/workers" % (self.address, cluster_name)
        req = HTTPRequest(
            url=url,
            method="PUT",
            body=json.dumps({"worker_count": n}),
            headers=HTTPHeaders({"Content-type": "application/json"}),
        )
        resp = await self._fetch(req)
        msg = json.loads(resp.body)
        if msg["message"]:
            warnings.warn(GatewayWarning(msg["message"]))

    def scale_cluster(self, cluster_name, n, **kwargs):
        """Scale a cluster to n workers.

        Parameters
        ----------
        cluster_name : str
            The cluster name.
        n : int
            The number of workers to scale to.
        """
        return self.sync(self._scale_cluster, cluster_name, n, **kwargs)


_widget_status_template = """
<div>
<style scoped>
    .dataframe tbody tr th:only-of-type {
        vertical-align: middle;
    }

    .dataframe tbody tr th {
        vertical-align: top;
    }

    .dataframe thead th {
        text-align: right;
    }
</style>
<table style="text-align: right;">
    <tr><th>Workers</th> <td>%d</td></tr>
    <tr><th>Cores</th> <td>%d</td></tr>
    <tr><th>Memory</th> <td>%s</td></tr>
</table>
</div>
"""


class GatewayCluster(object):
    """A dask-gateway cluster.

    Users should use ``Gateway.new_cluster`` or ``Gateway.connect`` instead of
    calling this class directly.
    """

    def __init__(self, gateway, name, scheduler_address, dashboard_link, security):
        self._gateway = gateway
        self.name = name
        self.scheduler_address = scheduler_address
        self.dashboard_link = dashboard_link
        self.security = security
        self._internal_client = None
        # XXX: Ensure client is closed. Currently disconnected clients try to
        # reconnect forever, which spams the schedulerproxy. Once this bug is
        # fixed, we should be able to remove this.
        self._clients = weakref.WeakSet()

    def __repr__(self):
        return "GatewayCluster<%s>" % self.name

    async def __aenter__(self):
        return self

    async def __aexit__(self):
        await self.shutdown()

    def __enter__(self):
        return self

    def __exit__(self):
        self.shutdown()

    def shutdown(self, **kwargs):
        """Shutdown this cluster."""
        if self._internal_client is not None:
            self._internal_client.close()
            self._internal_client = None
        for client in list(self._clients):
            client.close()
            self._clients.discard(client)
        return self._gateway.stop_cluster(self.name, **kwargs)

    def close(self, **kwargs):
        """Shutdown this cluster, alias for ``shutdown``"""
        return self.shutdown(**kwargs)

    def get_client(self, set_as_default=True):
        """Get a ``Client`` for this cluster.

        Returns
        -------
        client : dask.distributed.Client
        """
        client = Client(
            self.scheduler_address,
            security=self.security,
            set_as_default=set_as_default,
            asynchronous=self._gateway.asynchronous,
            loop=self._gateway.loop,
        )
        if not self._gateway.asynchronous:
            self._clients.add(client)
        return client

    def scale(self, n, **kwargs):
        """Scale the cluster to ``n`` workers.

        Parameters
        ----------
        n : int
            The number of workers to scale to.
        """
        return self._gateway.scale_cluster(self.name, n, **kwargs)

    def _widget_status(self):
        if self._internal_client is None:
            return None
        try:
            workers = self._internal_client._scheduler_identity["workers"]
        except KeyError:
            if self._internal_client.status in ("closing", "closed"):
                return None
        else:
            n_workers = len(workers)
            cores = sum(w["nthreads"] for w in workers.values())
            memory = sum(w["memory_limit"] for w in workers.values())

            return _widget_status_template % (n_workers, cores, format_bytes(memory))

    async def _widget_updater(self, widget):
        while True:
            value = self._widget_status()
            if value is None:
                break
            widget.value = value
            await gen.sleep(2)

    def _widget(self):
        """ Create IPython widget for display within a notebook """
        try:
            return self._cached_widget
        except AttributeError:
            pass

        if self._gateway.asynchronous:
            return None

        try:
            from ipywidgets import Layout, VBox, HBox, IntText, Button, HTML
        except ImportError:
            self._cached_widget = None
            return None

        try:
            self._internal_client = self.get_client(set_as_default=False)
        except Exception:
            return None

        layout = Layout(width="150px")

        title = HTML("<h2>GatewayCluster</h2>")

        status = HTML(self._widget_status(), layout=Layout(min_width="150px"))

        request = IntText(0, description="Workers", layout=layout)
        scale = Button(description="Scale", layout=layout)

        @scale.on_click
        def scale_cb(b):
            with log_errors():
                self.scale(request.value)

        elements = [title, HBox([status, request, scale])]

        if self.dashboard_link is not None:
            link = HTML(
                '<p><b>Dashboard: </b><a href="{0}" target="_blank">{0}'
                "</a></p>\n".format(self.dashboard_link)
            )
            elements.append(link)

        self._cached_widget = box = VBox(elements)

        self._internal_client.loop.add_callback(self._widget_updater, status)

        return box

    def _ipython_display_(self, **kwargs):
        widget = self._widget()
        if widget is not None:
            return widget._ipython_display_(**kwargs)
        else:
            from IPython.display import display

            data = {"text/plain": repr(self), "text/html": self._repr_html_()}
            display(data, raw=True)

    def _repr_html_(self):
        if self.dashboard_link is not None:
            dashboard = "<a href='{0}' target='_blank'>{0}</a>".format(
                self.dashboard_link
            )
        else:
            dashboard = "Not Available"
        return (
            "<div style='background-color: #f2f2f2; display: inline-block; "
            "padding: 10px; border: 1px solid #999999;'>\n"
            "  <h3>GatewayCluster</h3>\n"
            "  <ul>\n"
            "    <li><b>Name: </b>{name}\n"
            "    <li><b>Dashboard: </b>{dashboard}\n"
            "  </ul>\n"
            "</div>\n"
        ).format(name=self.name, dashboard=dashboard)



import os
from setuptools import setup, find_packages

import versioneer

install_requires = ["tornado", "dask>=2.2.0", "distributed>=2.2.0"]

extras_require = {
    "kerberos": [
        'pykerberos;platform_system!="Windows"',
        'winkerberos;platform_system=="Windows"',
    ]
}

setup(
    name="dask-gateway",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    maintainer="Jim Crist",
    maintainer_email="jiminy.crist@gmail.com",
    license="BSD",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: System :: Distributed Computing",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords="dask hadoop kubernetes HPC distributed cluster",
    description="A client library for interacting with a dask-gateway server",
    long_description=(
        open("README.rst").read() if os.path.exists("README.rst") else ""
    ),
    url="https://gateway.dask.org/",
    project_urls={
        "Documentation": "https://gateway.dask.org/",
        "Source": "https://github.com/dask/dask-gateway/",
        "Issue Tracker": "https://github.com/dask/dask-gateway/issues",
    },
    packages=find_packages(),
    package_data={"dask_gateway": ["*.yaml"]},
    install_requires=install_requires,
    extras_require=extras_require,
    entry_points={
        "console_scripts": [
            "dask-gateway-scheduler = dask_gateway.dask_cli:scheduler",
            "dask-gateway-worker = dask_gateway.dask_cli:worker",
        ]
    },
    zip_safe=False,
)



from __future__ import print_function, division, absolute_import

import argparse
import json
import logging
import os
import sys
from urllib.parse import urlparse, quote

from tornado import web
from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from tornado.ioloop import IOLoop, TimeoutError

from distributed import Scheduler, Worker, Nanny
from distributed.security import Security
from distributed.utils import ignoring
from distributed.cli.utils import install_signal_handlers
from distributed.proctitle import (
    enable_proctitle_on_children,
    enable_proctitle_on_current,
)
from distributed.diagnostics.plugin import SchedulerPlugin

from . import __version__ as VERSION


logger = logging.getLogger(__name__)


class BaseHandler(web.RequestHandler):
    @property
    def gateway_service(self):
        return self.settings.get("gateway_service")

    @property
    def auth_token(self):
        return self.settings.get("auth_token")

    def prepare(self):
        if self.request.headers.get("Content-Type", "").startswith("application/json"):
            self.json_data = json.loads(self.request.body)
        else:
            self.json_data = None

    def get_current_user(self):
        auth_header = self.request.headers.get("Authorization")
        if auth_header:
            auth_type, auth_token = auth_header.split(" ", 1)
            if auth_type == "token" and auth_token == self.auth_token:
                return "gateway"
        return None


class ScaleDownHandler(BaseHandler):
    @web.authenticated
    async def post(self):
        try:
            remove_count = self.json_data["remove_count"]
        except (TypeError, KeyError):
            raise web.HTTPError(405)
        result = await self.gateway_service.scale_down(remove_count)
        self.write(result)
        self.set_status(201)


class StatusHandler(BaseHandler):
    @web.authenticated
    async def get(self):
        result = self.gateway_service.status()
        self.write(result)
        self.set_status(201)


class GatewaySchedulerService(object):
    def __init__(self, scheduler, io_loop=None, plugin=None):
        self.scheduler = scheduler
        self.loop = io_loop or scheduler.loop
        routes = [("/api/scale_down", ScaleDownHandler), ("/api/status", StatusHandler)]
        self.app = web.Application(
            routes, gateway_service=self, auth_token=plugin.gateway.token
        )
        self.server = None
        self.plugin = plugin

    def listen(self, address):
        ip, port = address
        self.server = self.app.listen(address=ip, port=port)
        ports = {s.getsockname()[1] for s in self.server._sockets.values()}
        assert len(ports) == 1, "Only a single port allowed"
        self.port = ports.pop()

    def stop(self):
        if self.server is not None:
            self.server.stop()
            self.server = None

    async def scale_down(self, remove_count):
        n_workers = len(self.scheduler.workers)
        n_to_close = min(n_workers, remove_count)
        logger.info(
            "Scale down [remove_count: %d, n_workers: %d]", remove_count, n_workers
        )
        if n_to_close > 0:
            to_close = self.scheduler.workers_to_close(n=n_to_close)
            self.plugin.shutdown_requested.update(to_close)
            closed = await self.scheduler.retire_workers(
                workers=to_close, remove=True, close_workers=True
            )
            logger.info(
                "Closed %d workers, of %d suggested", len(closed), len(to_close)
            )
            workers_closed = [v["name"] for v in closed.values()]
        else:
            workers_closed = []
        return {"workers_closed": workers_closed, "n_workers": n_workers}

    def status(self):
        workers = [
            ws.name for ws in self.scheduler.workers.values() if ws.status != "closed"
        ]
        return {"workers": workers}


class GatewaySchedulerPlugin(SchedulerPlugin):
    """A plugin to notify the gateway when workers are added or removed"""

    def __init__(self, gateway, loop):
        self.gateway = gateway
        # A mapping from name to WorkerState. All workers have a specified
        # unique name that will be consistent between worker restarts. The
        # worker address may not be consistent though, so we track intermittent
        # connection failures from the worker name only.
        self.workers = {}
        # A set of worker *addresses* whose shutdown is requested
        self.shutdown_requested = set()
        self.timeouts = {}
        self.loop = loop

    def add_worker(self, scheduler, worker):
        ws = scheduler.workers[worker]
        logger.debug("Worker added [address: %r, name: %r]", ws.address, ws.name)
        timeout = self.timeouts.pop(ws.name, None)
        if timeout is not None:
            # Existing timeout running for this worker, cancel it
            self.loop.remove_timeout(timeout)
        self.workers[worker] = ws
        self.loop.add_callback(self.gateway.notify_worker_added, ws)

    def remove_worker(self, scheduler, worker):
        ws = self.workers.pop(worker)
        logger.debug("Worker removed [address: %r, name: %r]", ws.address, ws.name)

        try:
            self.shutdown_requested.remove(ws.address)
            # This worker was expected to exit, no need to notify
            return
        except KeyError:
            # Unexpected worker shutdown
            pass

        # Start a timer to notify the gateway that the worker is permanently
        # gone. This allows for short-lived communication failures, while
        # still detecting worker failures.
        async def callback():
            logger.debug("Notifying worker %s removed", ws.name)
            await self.gateway.notify_worker_removed(ws)
            self.timeouts.pop(ws.name, None)

        self.timeouts[ws.name] = self.loop.call_later(5, callback)


class GatewayClient(object):
    def __init__(self, cluster_name, api_token, api_url):
        self.cluster_name = cluster_name
        self.token = api_token
        self.api_url = api_url

    async def send_addresses(self, scheduler, dashboard, api):
        client = AsyncHTTPClient()
        body = json.dumps(
            {
                "scheduler_address": scheduler,
                "dashboard_address": dashboard,
                "api_address": api,
            }
        )
        url = "%s/clusters/%s/addresses" % (self.api_url, self.cluster_name)
        req = HTTPRequest(
            url,
            method="PUT",
            headers={
                "Authorization": "token %s" % self.token,
                "Content-type": "application/json",
            },
            body=body,
        )
        await client.fetch(req)

    async def get_scheduler_address(self):
        client = AsyncHTTPClient()
        url = "%s/clusters/%s/addresses" % (self.api_url, self.cluster_name)
        req = HTTPRequest(
            url, method="GET", headers={"Authorization": "token %s" % self.token}
        )
        resp = await client.fetch(req)
        data = json.loads(resp.body.decode("utf8", "replace"))
        return data["scheduler_address"]

    async def notify_worker_added(self, ws):
        client = AsyncHTTPClient()
        body = json.dumps({"name": ws.name, "address": ws.address})
        url = "%s/clusters/%s/workers/%s" % (
            self.api_url,
            self.cluster_name,
            quote(ws.name),
        )
        req = HTTPRequest(
            url,
            method="PUT",
            headers={
                "Authorization": "token %s" % self.token,
                "Content-type": "application/json",
            },
            body=body,
        )
        await client.fetch(req)

    async def notify_worker_removed(self, ws):
        client = AsyncHTTPClient()
        url = "%s/clusters/%s/workers/%s" % (
            self.api_url,
            self.cluster_name,
            quote(ws.name),
        )
        req = HTTPRequest(
            url, method="DELETE", headers={"Authorization": "token %s" % self.token}
        )
        await client.fetch(req)


scheduler_parser = argparse.ArgumentParser(
    prog="dask-gateway-scheduler", description="Start a dask-gateway scheduler"
)
scheduler_parser.add_argument("--version", action="version", version=VERSION)


def getenv(key):
    out = os.environ.get(key)
    if not out:
        logger.error("Environment variable %r not found, shutting down", key)
        sys.exit(1)
    return out


def make_security(tls_cert=None, tls_key=None):
    tls_cert = tls_cert or getenv("DASK_GATEWAY_TLS_CERT")
    tls_key = tls_key or getenv("DASK_GATEWAY_TLS_KEY")

    return Security(
        tls_ca_file=tls_cert,
        tls_scheduler_cert=tls_cert,
        tls_scheduler_key=tls_key,
        tls_worker_cert=tls_cert,
        tls_worker_key=tls_key,
    )


def make_gateway_client(cluster_name=None, api_url=None, api_token=None):
    cluster_name = cluster_name or getenv("DASK_GATEWAY_CLUSTER_NAME")
    api_url = api_url or getenv("DASK_GATEWAY_API_URL")
    api_token = api_token or getenv("DASK_GATEWAY_API_TOKEN")
    return GatewayClient(cluster_name, api_token, api_url)


def scheduler(argv=None):
    scheduler_parser.parse_args(argv)

    gateway = make_gateway_client()
    security = make_security()

    loop = IOLoop.current()

    install_signal_handlers(loop)
    enable_proctitle_on_current()
    enable_proctitle_on_children()

    if sys.platform.startswith("linux"):
        import resource  # module fails importing on Windows

        soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        limit = max(soft, hard // 2)
        resource.setrlimit(resource.RLIMIT_NOFILE, (limit, hard))

    async def run():
        scheduler = await start_scheduler(gateway, security)
        await scheduler.finished()

    try:
        loop.run_sync(run)
    except (KeyboardInterrupt, TimeoutError):
        pass


async def start_scheduler(gateway, security, exit_on_failure=True):
    loop = IOLoop.current()
    plugin = GatewaySchedulerPlugin(gateway, loop)

    services = {("gateway", 0): (GatewaySchedulerService, {"plugin": plugin})}
    dashboard = False
    with ignoring(ImportError):
        from distributed.dashboard.scheduler import BokehScheduler

        services[("dashboard", 0)] = (BokehScheduler, {})
        dashboard = True

    scheduler = Scheduler(loop=loop, services=services, security=security)
    scheduler.add_plugin(plugin)
    await scheduler

    host = urlparse(scheduler.address).hostname
    gateway_port = scheduler.services["gateway"].port
    api_address = "http://%s:%d" % (host, gateway_port)

    if dashboard:
        dashboard_port = scheduler.services["dashboard"].port
        dashboard_address = "http://%s:%d" % (host, dashboard_port)
    else:
        dashboard_address = ""

    try:
        await gateway.send_addresses(scheduler.address, dashboard_address, api_address)
    except Exception as exc:
        logger.error("Failed to send addresses to gateway", exc_info=exc)
        if exit_on_failure:
            sys.exit(1)

    return scheduler


worker_parser = argparse.ArgumentParser(
    prog="dask-gateway-worker", description="Start a dask-gateway worker"
)

worker_parser.add_argument("--version", action="version", version=VERSION)
worker_parser.add_argument(
    "--nthreads", type=int, default=1, help="The number of threads to use"
)
worker_parser.add_argument(
    "--memory-limit", default="auto", help="The maximum amount of memory to allow"
)
worker_parser.add_argument("--name", default=None, help="The worker name")


async def start_worker(
    gateway,
    security,
    worker_name,
    nthreads=1,
    memory_limit="auto",
    local_directory="",
    nanny=True,
):
    loop = IOLoop.current()

    scheduler = await gateway.get_scheduler_address()

    typ = Nanny if nanny else Worker

    worker = typ(
        scheduler,
        loop=loop,
        nthreads=nthreads,
        memory_limit=memory_limit,
        security=security,
        name=worker_name,
        local_directory=local_directory,
    )

    if nanny:

        async def close(signalnum):
            await worker.close(timeout=2)

        install_signal_handlers(loop, cleanup=close)

    await worker
    return worker


def worker(argv=None):
    args = worker_parser.parse_args(argv)

    worker_name = args.name or getenv("DASK_GATEWAY_WORKER_NAME")
    nthreads = args.nthreads
    memory_limit = args.memory_limit

    gateway = make_gateway_client()
    security = make_security()

    enable_proctitle_on_current()
    enable_proctitle_on_children()

    loop = IOLoop.current()

    async def run():
        worker = await start_worker(
            gateway, security, worker_name, nthreads, memory_limit
        )
        await worker.finished()

    try:
        loop.run_sync(run)
    except (KeyboardInterrupt, TimeoutError):
        pass



# This code was borrowed and modified from aiohttp to work with tornado. The
# original source can be found here:
#
# https://github.com/aio-libs/aiohttp/blob/master/aiohttp/cookiejar.py
#
# The original aiohttp code is licensed as follows:
#
# Copyright 2013-2019 Nikolay Kim and Andrew Svetlov
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
import datetime
import re
from collections import defaultdict
from http.cookies import SimpleCookie
from math import ceil
from urllib.parse import urlparse

from tornado.gen import IOLoop


_DATE_TOKENS_RE = re.compile(
    r"[\x09\x20-\x2F\x3B-\x40\x5B-\x60\x7B-\x7E]*"
    r"(?P<token>[\x00-\x08\x0A-\x1F\d:a-zA-Z\x7F-\xFF]+)"
)

_DATE_HMS_TIME_RE = re.compile(r"(\d{1,2}):(\d{1,2}):(\d{1,2})")

_DATE_DAY_OF_MONTH_RE = re.compile(r"(\d{1,2})")

_DATE_MONTH_RE = re.compile(
    "(jan)|(feb)|(mar)|(apr)|(may)|(jun)|(jul)|" "(aug)|(sep)|(oct)|(nov)|(dec)", re.I
)

_DATE_YEAR_RE = re.compile(r"(\d{2,4})")


def _is_domain_match(domain, hostname):
    """Implements domain matching adhering to RFC 6265."""
    if hostname == domain:
        return True

    if not hostname.endswith(domain):
        return False

    return hostname[: -len(domain)].endswith(".")


def _is_path_match(req_path, cookie_path):
    """Implements path matching adhering to RFC 6265."""
    if not req_path.startswith("/"):
        req_path = "/"

    if req_path == cookie_path:
        return True

    if not req_path.startswith(cookie_path):
        return False

    if cookie_path.endswith("/"):
        return True

    return req_path[len(cookie_path) :].startswith("/")


def _parse_date(date_str):
    """Implements date string parsing adhering to RFC 6265."""
    if not date_str:
        return None

    found_time = False
    found_day = False
    found_month = False
    found_year = False

    hour = minute = second = 0
    day = 0
    month = 0
    year = 0

    for token_match in _DATE_TOKENS_RE.finditer(date_str):

        token = token_match.group("token")

        if not found_time:
            time_match = _DATE_HMS_TIME_RE.match(token)
            if time_match:
                found_time = True
                hour, minute, second = [int(s) for s in time_match.groups()]
                continue

        if not found_day:
            day_match = _DATE_DAY_OF_MONTH_RE.match(token)
            if day_match:
                found_day = True
                day = int(day_match.group())
                continue

        if not found_month:
            month_match = _DATE_MONTH_RE.match(token)
            if month_match:
                found_month = True
                month = month_match.lastindex
                continue

        if not found_year:
            year_match = _DATE_YEAR_RE.match(token)
            if year_match:
                found_year = True
                year = int(year_match.group())

    if 70 <= year <= 99:
        year += 1900
    elif 0 <= year <= 69:
        year += 2000

    if False in (found_day, found_month, found_year, found_time):
        return None

    if not 1 <= day <= 31:
        return None

    if year < 1601 or hour > 23 or minute > 59 or second > 59:
        return None

    return datetime.datetime(
        year, month, day, hour, minute, second, tzinfo=datetime.timezone.utc
    )


class CookieJar(object):
    """Implements cookie storage adhering to RFC 6265."""

    _MAX_TIME = 2051215261.0  # so far in future (2035-01-01)

    def __init__(self, loop=None):
        self._loop = loop or IOLoop.current()
        self._cookies = defaultdict(SimpleCookie)
        self._host_only_cookies = set()
        self._next_expiration = ceil(self._loop.time())
        self._expirations = {}

    def clear(self):
        self._cookies.clear()
        self._host_only_cookies.clear()
        self._next_expiration = ceil(self._loop.time())
        self._expirations.clear()

    def __iter__(self):
        self._do_expiration()
        return (m for c in self._cookies.values() for m in c.values())

    def __len__(self):
        self._do_expiration()
        return sum(map(len, self._cookies.values()))

    def _do_expiration(self):
        now = self._loop.time()
        if self._next_expiration > now:
            return
        if not self._expirations:
            return
        next_expiration = self._MAX_TIME
        to_del = []
        cookies = self._cookies
        expirations = self._expirations
        for (domain, name), when in expirations.items():
            if when <= now:
                cookies[domain].pop(name, None)
                to_del.append((domain, name))
                self._host_only_cookies.discard((domain, name))
            else:
                next_expiration = min(next_expiration, when)
        for key in to_del:
            del expirations[key]

        self._next_expiration = ceil(next_expiration)

    def _expire_cookie(self, when, domain, name):
        iwhen = int(when)
        self._next_expiration = min(self._next_expiration, iwhen)
        self._expirations[(domain, name)] = iwhen

    def update_cookies(self, cookies, effective_url=None):
        """Update cookies."""
        if effective_url:
            effective_url = urlparse(effective_url)
            hostname = effective_url.hostname
            url_path = effective_url.path
        else:
            hostname = None
            url_path = "/"

        if isinstance(cookies, str):
            cookies = SimpleCookie(cookies)

        for name, cookie in cookies.items():
            domain = cookie["domain"]

            # ignore domains with trailing dots
            if domain.endswith("."):
                domain = ""
                del cookie["domain"]

            if not domain and hostname is not None:
                # Set the cookie's domain to the response hostname
                # and set its host-only-flag
                self._host_only_cookies.add((hostname, name))
                domain = cookie["domain"] = hostname

            if domain.startswith("."):
                # Remove leading dot
                domain = domain[1:]
                cookie["domain"] = domain

            if hostname and not _is_domain_match(domain, hostname):
                # Setting cookies for different domains is not allowed
                continue

            path = cookie["path"]
            if not path or not path.startswith("/"):
                # Set the cookie's path to the response path
                path = url_path
                if not path.startswith("/"):
                    path = "/"
                else:
                    # Cut everything from the last slash to the end
                    path = "/" + path[1 : path.rfind("/")]
                cookie["path"] = path

            max_age = cookie["max-age"]
            if max_age:
                try:
                    delta_seconds = int(max_age)
                    self._expire_cookie(self._loop.time() + delta_seconds, domain, name)
                except ValueError:
                    cookie["max-age"] = ""

            else:
                expires = cookie["expires"]
                if expires:
                    expire_time = _parse_date(expires)
                    if expire_time:
                        self._expire_cookie(expire_time.timestamp(), domain, name)
                    else:
                        cookie["expires"] = ""

            self._cookies[domain][name] = cookie

        self._do_expiration()

    def filter_cookies(self, request_url):
        """Returns this jar's cookies filtered for a request url."""
        self._do_expiration()
        request_url = urlparse(request_url)
        hostname = request_url.hostname or ""
        filtered = SimpleCookie()
        is_not_secure = request_url.scheme not in ("https", "wss")

        for cookie in self._cookies.values():
            for morsel in cookie.values():
                name = morsel.key
                domain = morsel["domain"]

                # Send shared cookies
                if not domain:
                    filtered[name] = morsel
                    continue

                if (domain, name) in self._host_only_cookies:
                    if domain != hostname:
                        continue
                elif not _is_domain_match(domain, hostname):
                    continue

                if not _is_path_match(request_url.path, morsel["path"]):
                    continue

                if is_not_secure and morsel["secure"]:
                    continue

                filtered[name] = morsel

        return filtered

    def pre_request(self, req):
        """Update a HTTPRequest with any relevant cookies"""
        cookies = self.filter_cookies(req.url)
        for v in cookies.values():
            req.headers.add("Cookie", v.OutputString(attrs=()))

    def post_response(self, resp):
        """Update our cookiejar with the results from a HTTPResponse"""
        new = resp.headers.get_list("Set-Cookie")
        if new:
            cookies = SimpleCookie()
            for m in new:
                cookies.load(m)
            self.update_cookies(cookies, resp.effective_url)



import getpass
import os
import re
from base64 import b64encode
from urllib.parse import urlparse

import dask


__all__ = ("GatewayAuth", "BasicAuth", "KerberosAuth", "JupyterHubAuth", "get_auth")


def _import_object(name):
    parts = name.rsplit(".", 1)
    if len(parts) == 2:
        package, obj = parts
        mod = __import__(package, fromlist=[obj])
        try:
            return getattr(mod, obj)
        except AttributeError:
            raise ImportError("Failed to import %s" % name)
    else:
        return __import__(name)


def get_auth(auth=None):
    """Get a ``GatewayAuth`` instance.

    Creates an authenticator depending on the given parameters and gateway
    configuration.
    """
    if isinstance(auth, GatewayAuth):
        return auth

    if auth is None:
        auth = dask.config.get("gateway.auth.type", None)

    if isinstance(auth, str):
        if auth == "kerberos":
            auth = KerberosAuth
        elif auth == "basic":
            auth = BasicAuth
        elif auth == "jupyterhub":
            auth = JupyterHubAuth
        else:
            auth = _import_object(auth)
    elif not callable(auth):
        raise TypeError("Unknown auth value %r" % auth)

    auth_kwargs = dask.config.get("gateway.auth.kwargs", None) or {}
    out = auth(**auth_kwargs)

    if not isinstance(out, GatewayAuth):
        raise TypeError("auth must be instance of GatewayAuth, got %r" % out)
    return out


class GatewayAuth(object):
    """Base class for authenticating clients in dask-gateway"""

    def __init__(self, **kwargs):
        pass

    def pre_request(self, req, resp):
        """Authentication hook before client request is sent.

        This is only called after a 401 response from the server.

        Parameters
        ----------
        req : HTTPRequest
            The request to authenticate
        resp : HTTPResponse
            The 401 HTTPResponse that triggered the authentication step.

        Returns
        -------
        context : object
            If necessary, may optionally return an opaque object storing
            additional context needed to complete the authentication process.
            This will be passed to ``post_response``.
        """
        pass

    def post_response(self, req, resp, context=None):
        """Authentication hook after server response is received.

        Parameters
        ----------
        req : HTTPRequest
            The submitted HTTPRequest
        resp : HTTPResponse
            The server response
        context : object or None
            The return value of ``pre_request``, providing any additional
            context needed to complete the authentication process.
        """
        pass


class BasicAuth(GatewayAuth):
    """Attaches HTTP Basic Authentication to the given Request object."""

    def __init__(self, username=None, password=None):
        if username is None:
            username = getpass.getuser()
        if password is None:
            password = ""
        self.username = username
        self.password = password

    def pre_request(self, req, resp):
        data = b":".join(
            (self.username.encode("latin1"), self.password.encode("latin1"))
        )
        req.headers["Authorization"] = "Basic " + b64encode(data).decode()


class KerberosAuth(GatewayAuth):
    """Authenticate with kerberos"""

    auth_regex = re.compile(r"(?:.*,)*\s*Negotiate\s*([^,]*),?", re.I)  # noqa

    def pre_request(self, req, resp):
        # TODO: convert errors to some common error class
        import kerberos

        hostname = urlparse(resp.effective_url).hostname
        _, context = kerberos.authGSSClientInit(
            "HTTP@%s" % hostname,
            gssflags=kerberos.GSS_C_MUTUAL_FLAG | kerberos.GSS_C_SEQUENCE_FLAG,
        )
        kerberos.authGSSClientStep(context, "")
        response = kerberos.authGSSClientResponse(context)
        req.headers["Authorization"] = "Negotiate " + response
        return context

    def post_response(self, req, resp, context):
        import kerberos

        www_auth = resp.headers.get("www-authenticate", None)
        token = None
        if www_auth:
            match = self.auth_regex.search(www_auth)
            if match:
                token = match.group(1)
        if not token:
            raise Exception("Kerberos negotiation failed")
        kerberos.authGSSClientStep(context, token)


class JupyterHubAuth(GatewayAuth):
    """Uses JupyterHub API tokens to authenticate"""

    def __init__(self, api_token=None):
        if api_token is None:
            api_token = os.environ.get("JUPYTERHUB_API_TOKEN")
            if api_token is None:
                raise ValueError(
                    "No JupyterHub API token provided, and JUPYTERHUB_API_TOKEN "
                    "environment variable not found"
                )
        self.api_token = api_token

    def pre_request(self, req, resp):
        req.headers["Authorization"] = "jupyterhub " + self.api_token
